@@ -99,7 +99,7 @@ class GJOChallengeLoader(ChallengeLoader):
         problem_id_to_correct_idx = {}
 
         for _, row in filtered_predictions_df.iterrows():
-            problem_id: int = int(row['problem_id'])
+            problem_id: str = str(row['problem_id'])
             username: str = str(row['username'])
             # the original timestamp is in string format like "2024-09-10T19:22:23Z"
             timestamp: datetime = datetime.fromisoformat(str(row['timestamp']))
@@ -113,11 +113,12 @@ class GJOChallengeLoader(ChallengeLoader):
                     problem_meta_row['correct_answer'])
 
             forecast_event = ForecastEvent(
+                forecast_id=f"{problem_id}-{username}",
                 problem_id=problem_id,
                 username=username,
                 timestamp=timestamp,
                 probs=probs,
-                correct_prob=probs[problem_id_to_correct_idx[problem_id]]
+                unnormalized_probs=probs,
             )
 
             problem_id_to_forecast_events[problem_id].append(forecast_event)
@@ -125,15 +126,15 @@ class GJOChallengeLoader(ChallengeLoader):
         # Iterate over each row of the filtered metadata df to construct the forecast problems
         forecast_problems = []
         for _, row in filtered_metadata_df.iterrows():
-            problem_id: int = int(row['problem_id'])
+            problem_id: str = str(row['problem_id'])
             problem_forecasts = problem_id_to_forecast_events[problem_id]
             forecast_problems.append(ForecastProblem(
                 title=str(row['title']),
                 problem_id=problem_id,
                 options=list(row['options']),
-                correct_option=str(row['correct_answer']),
+                correct_option_idx=problem_id_to_correct_idx[problem_id],
                 forecasts=problem_forecasts,
-                end_date=datetime.fromisoformat(
+                end_time=datetime.fromisoformat(
                     str(row['metadata']['end_date'])),
                 num_forecasters=len(problem_forecasts),
                 url=str(row['url']),
@@ -173,7 +174,7 @@ class ProphetArenaChallengeLoader(ChallengeLoader):
     """Load forecast challenges from Prophet Arena data format."""
 
     def __init__(self, predictions_df: Optional[pd.DataFrame] = None, predictions_file: Optional[str] = None,
-                 challenge_title: str = "", use_bid_for_odds: bool = False):
+                 challenge_title: str = "", use_bid_for_odds: bool = False, use_open_time: bool = False):
         """Initialize the ProphetArenaChallengeLoader.
 
         The challenge can be either loaded with a given `pd.DataFrame` or with a path to a predictions file.
@@ -184,9 +185,13 @@ class ProphetArenaChallengeLoader(ChallengeLoader):
         :param use_bid_for_odds: Whether to use the `yes_bid` field for implied probability calculation.
                                 If True, the implied probability will be calculated as the (yes_bid + no_bid) / 2.
                                 If False, the implied probability will be simply `yes_ask` (normalized to sum to 1).
+        :param use_open_time: Whether to use the `open_time` field for the `end_time` of the problem.
+                                If True, the `end_time` will be the `open_time` of the problem.
+                                If False, the `end_time` will be the `close_time` of the problem.
         """
         self.challenge_title = challenge_title
         self.use_bid_for_odds = use_bid_for_odds
+        self.use_open_time = use_open_time
         self.logger = get_logger(
             f"pm_rank.data.loaders.{self.__class__.__name__}")
         if predictions_df is None:
@@ -202,42 +207,59 @@ class ProphetArenaChallengeLoader(ChallengeLoader):
             self.logger.info(f"Initialize challenge loader with pd.DataFrame")
 
     @staticmethod
-    def _calculate_implied_probs_for_problem(market_info: dict, options: list, use_bid_for_odds: bool = False, logger: Optional[logging.Logger] = None) -> list | None:
+    def _calculate_implied_probs_for_problem(market_info: dict, options: list, use_bid_for_odds: bool = False, \
+        yes_contract: bool = True, logger: Optional[logging.Logger] = None) -> List:
         """
         Calculate odds for each option from market_info dict.
         For multi-option, use yes_ask for each option and normalize to sum to 1 (implied probabilities).
         """
+        ask_str = 'yes_ask' if yes_contract else 'no_ask'
+        bid_str = 'yes_bid' if yes_contract else 'no_bid'
+
         asks = []
         for opt in options:
             info = market_info.get(opt, {})
-            yes_ask = info.get('yes_ask', None)
-            if yes_ask is not None and yes_ask > 0:
+            yes_ask = info.get(ask_str, None)
+            if info.get('liquidity', None) is not None and info.get('liquidity', None) < 100:
+                asks.append(100)
+            elif yes_ask is not None and yes_ask > 0:
                 if use_bid_for_odds:
-                    if 'yes_bid' in info:
-                        yes_bid = info['yes_bid']
+                    if bid_str in info:
+                        yes_bid = info[bid_str]
                         asks.append((yes_bid + yes_ask) / 2)
+                        # asks.append(max(yes_bid, yes_ask))
                     else:
                         asks.append(yes_ask)
                 else:
                     asks.append(yes_ask)
             else:
-                warning_msg = f"Warning: Option {opt} in market {market_info.get('title', 'Unknown Market')} has no odds info"
+                warning_msg = f"Warning: Option {opt} in market {market_info.get('title', 'Unknown Market')} does not have odds info"
                 logger.warning(warning_msg) if logger is not None else print(
                     warning_msg)
                 asks.append(None)
-        # normalize the implied probabilities to sum to 1
-        implied_probs = [(a / 100.0) if a is not None else 0.0 for a in asks]
-        total = sum(implied_probs)
 
-        if total > 0:
-            return [p / total for p in implied_probs]
-        else:
-            return None
+        implied_probs = [(a / 100.0) if a is not None else 1e-3 for a in asks]
+        return implied_probs
 
-    def load_challenge(self) -> ForecastChallenge:
+    @staticmethod
+    def _get_normalized_probs(unnormalized_probs: list) -> list:
+        """
+        Get normalized probabilities from unnormalized probabilities.
+        """
+        if not math.isclose(sum(unnormalized_probs), 1.0, abs_tol=1e-6):
+            if sum(unnormalized_probs) > 0:
+                return [p / sum(unnormalized_probs) for p in unnormalized_probs]
+            else:
+                return [1.0 / len(unnormalized_probs) for _ in unnormalized_probs]
+        return unnormalized_probs
+
+    def load_challenge(self, add_market_baseline: bool = False) -> ForecastChallenge:
         """
         Load challenge data from Prophet Arena data format.
         Group by submission_id, then for each group, build the list of forecasts, then the ForecastProblem.
+
+        :param add_market_baseline: Whether to add the market baseline as a forecaster
+        :return: A ForecastChallenge object containing the forecast problems and events.
         """
         if hasattr(self, 'predictions_df'):
             df = self.predictions_df
@@ -245,78 +267,109 @@ class ProphetArenaChallengeLoader(ChallengeLoader):
             df = pd.read_csv(self.predictions_file)
 
         forecast_problems = []
-        problem_id_counter = 1
-        grouped = df.groupby('submission_id')
+        categories = []
+        grouped = df.groupby('event_ticker')
 
-        self.logger.info(
-            f"Currently, the Prophet Arena challenge is using the `open_time` in place of `close_time` for each problem.")
-        self.logger.info(
-            f"This is because the `close_time` are too far from each other, and the `open_time` is more consistent.")
+        if self.use_open_time:
+            self.logger.warning(
+                f"Currently, the Prophet Arena challenge is using the `open_time` in place of `close_time` for each problem.")
 
-        for submission_id, group in grouped:
+        for event_ticker, group in grouped:
             first_row = group.iloc[0]
-            options = parse_json_or_eval(
-                first_row['markets'], expect_type=list)
+            problem_id = str(event_ticker)
             market_info = parse_json_or_eval(
                 first_row['market_info'], expect_type=dict)
             # skip this market if the market_info is empty
             if not market_info:
                 continue
             first_option_info = next(iter(market_info.values()))
-            title = first_option_info.get('title', submission_id)
-            open_date_str = first_option_info.get('open_time', None)
-            open_date = datetime.fromisoformat(open_date_str.replace(
-                'Z', '+00:00')) if open_date_str else datetime.now()
+            title = first_option_info.get('title', event_ticker)
+
+            if self.use_open_time:
+                open_time = first_option_info.get('open_time', None)
+                end_time = datetime.fromisoformat(open_time.replace(
+                    'Z', '+00:00')) if open_time else datetime.now()
+            else:
+                if 'close_time' in first_row:
+                    close_time = first_row['close_time']
+                else:
+                    close_time = first_option_info.get('close_time', None)
+                end_time = datetime.fromisoformat(close_time.replace(
+                    'Z', '+00:00')) if close_time else datetime.now()
+            
+            problem_option_keys = list(market_info.keys())
+
             odds = self._calculate_implied_probs_for_problem(
-                market_info, options, self.use_bid_for_odds, self.logger)
+                market_info, problem_option_keys, self.use_bid_for_odds, True, self.logger)
+
+            no_odds = self._calculate_implied_probs_for_problem(
+                market_info, problem_option_keys, self.use_bid_for_odds, False, self.logger)
+
             market_outcome = parse_json_or_eval(
                 first_row['market_outcome'], expect_type=dict)
 
-            correct_option = [opt for opt, val in (
-                market_outcome or {}).items() if int(val) == 1][0]
-
-            if correct_option not in options:
-                continue
+            correct_option_idx = [i for i, key in enumerate(problem_option_keys) if market_outcome[key] == 1]
+            timestamp = datetime.now()
 
             forecasts = []
+
+            if add_market_baseline:
+                # we will add a "market row" in this group, with (unnormalized) prob being simply the market odds
+                forecasts.append(ForecastEvent(
+                    forecast_id=f"{problem_id}-market-baseline",
+                    problem_id=problem_id,
+                    username="market-baseline",
+                    timestamp=timestamp,
+                    probs=self._get_normalized_probs(odds),
+                    unnormalized_probs=odds,
+                ))
+
+            category = first_row.get('category', None)
+            if category is not None and category not in categories:
+                categories.append(category)
+
             for _, row in group.iterrows():
                 username = str(row['predictor_name'])
                 prediction: dict = parse_json_or_eval(
                     row['prediction'], expect_type=dict)
                 probs_dict = {d['market']: d['probability']
                               for d in prediction.get('probabilities', [])}
-                probs = [probs_dict.get(opt, 0.0) for opt in options]
+                unnormalized_probs = [probs_dict.get(opt, 0.0) for opt in problem_option_keys]
                 # make sure the probs sum to 1
-                if not math.isclose(sum(probs), 1.0, abs_tol=1e-6):
-                    continue
-                timestamp = datetime.now()
-                correct_prob = probs[options.index(
-                    correct_option)] if correct_option in options else 0.0
+                probs = self._get_normalized_probs(unnormalized_probs)
+
+                # set `forecast_id` to be `prediction_id` if the column exists, otherwise construct one
+                # using `username` and `problem_id`
+                forecast_id = str(row['prediction_id']) if 'prediction_id' in row else f"{username}-{problem_id}"
+
                 forecasts.append(ForecastEvent(
-                    problem_id=problem_id_counter,
+                    forecast_id=forecast_id,
+                    problem_id=problem_id,
                     username=username,
                     timestamp=timestamp,
                     probs=probs,
-                    correct_prob=correct_prob
+                    unnormalized_probs=unnormalized_probs,
                 ))
 
             if len(forecasts) > 0:
                 forecast_problems.append(ForecastProblem(
                     title=title,
-                    problem_id=problem_id_counter,
-                    options=options,
-                    correct_option=correct_option,
+                    problem_id=problem_id,
+                    options=problem_option_keys,
+                    correct_option_idx=correct_option_idx,
                     forecasts=forecasts,
-                    end_date=open_date,
+                    end_time=end_time,
                     num_forecasters=len(forecasts),
                     url=None,
-                    odds=odds
+                    odds=odds,
+                    no_odds=no_odds,
+                    category=category
                 ))
-                problem_id_counter += 1
 
         forecast_challenge = ForecastChallenge(
             title=self.challenge_title or "Prophet Arena Challenge",
-            forecast_problems=forecast_problems
+            forecast_problems=forecast_problems,
+            categories=categories
         )
         return forecast_challenge
 
