@@ -25,8 +25,9 @@ Key Concepts:
 
 from abc import ABC, abstractmethod
 import numpy as np
-from typing import List, Iterator, Dict, Tuple, Any
-from pm_rank.data.base import ForecastProblem
+from typing import List, Iterator, Dict, Tuple, Any, Literal, Callable
+from collections import OrderedDict
+from pm_rank.data.base import ForecastProblem, ForecastChallenge
 from pm_rank.model.utils import forecaster_data_to_rankings, get_logger, log_ranking_table
 import logging
 
@@ -57,14 +58,14 @@ class ScoringRule(ABC):
             self.logger.setLevel(logging.DEBUG)
 
     @abstractmethod
-    def _score_fn(self, correct_probs: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
+    def _score_fn(self, correct_option_idx: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
         """Implement the scoring function for the specific rule.
 
         This abstract method must be implemented by subclasses to define the
         specific mathematical formulation of the scoring rule.
 
-        :param correct_probs: Array of predicted probabilities for the actual outcomes.
-                             Shape (n,) where n is the number of forecasts.
+        :param correct_option_idx: Array of indices of the correct options.
+                             Shape (m,) where m is the number of correct options.
         :param all_probs: Array of all predicted probability distributions.
                          Shape (n, k) where n is number of forecasts, k is number of options.
 
@@ -98,8 +99,8 @@ class ScoringRule(ABC):
             problem_weights / np.sum(problem_weights)
         return problem_weights
 
-    def fit(self, problems: List[ForecastProblem], problem_discriminations: np.ndarray | List[float] | None = None, include_scores: bool = True) \
-            -> Tuple[Dict[str, Any], Dict[str, int]] | Dict[str, int]:
+    def fit(self, problems: List[ForecastProblem], problem_discriminations: np.ndarray | List[float] | None = None, include_scores: bool = True, \
+        include_per_problem_info: bool = False) -> Tuple[Dict[str, Any], Dict[str, int]] | Dict[str, int]:
         """Fit the scoring rule to the given problems and return rankings.
 
         This method processes all problems and calculates scores for each forecaster
@@ -111,10 +112,14 @@ class ScoringRule(ABC):
         :param problem_discriminations: Optional array of discrimination parameters for
                                        weighting problems. If None, all problems are weighted equally.
         :param include_scores: Whether to include scores in the results (default: True).
+        :param include_per_problem_info: Whether to include per-problem info in the results (default: False).
 
         :returns: Ranking results, either as a tuple of (scores, rankings) or just rankings.
+                  If include_per_problem_info is True, returns a tuple of (scores, rankings, per_problem_info).
         """
         forecaster_data = {}
+        if include_per_problem_info:
+            per_problem_info = []
 
         if problem_discriminations is not None:
             problem_weights = self._get_problem_weights(
@@ -123,29 +128,89 @@ class ScoringRule(ABC):
             problem_weights = np.ones(len(problems))
 
         for i, problem in enumerate(problems):
-            correct_probs, all_probs, usernames = [], [], []
+            all_probs, usernames = [], []
+            correct_option_idx = np.array(problem.correct_option_idx)
             for forecast in problem.forecasts:
                 username = forecast.username
                 if username not in forecaster_data:
                     forecaster_data[username] = []
                 usernames.append(username)
-                correct_probs.append(forecast.correct_prob)
-                all_probs.append(forecast.probs)
+                all_probs.append(forecast.unnormalized_probs)
 
-            correct_probs = np.array(correct_probs)
             all_probs = np.array(all_probs)
             # weight the scores by the problem weights
             scores = self._score_fn(
-                correct_probs, all_probs) * problem_weights[i]
+                correct_option_idx, all_probs) * problem_weights[i]
             # attribute the scores to the forecasters
             for username, score in zip(usernames, scores):
                 forecaster_data[username].append(score)
 
+            if include_per_problem_info:
+                for username, score in zip(usernames, scores):
+                    per_problem_info.append({
+                        "forecast_id": forecast.forecast_id,
+                        "username": username,
+                        "problem_title": problem.title,
+                        "problem_id": problem.problem_id,
+                        "problem_category": problem.category,
+                        "score": score,
+                        "probs": forecast.unnormalized_probs
+                    })
+
         result = forecaster_data_to_rankings(
             forecaster_data, include_scores=include_scores, ascending=False, aggregate="mean")
         if self.verbose:
             log_ranking_table(self.logger, result)
-        return result
+        return (*result, per_problem_info) if include_per_problem_info else result
+
+    def _fit_stream_generic(self, batch_iter: Iterator, key_fn: Callable, include_scores: bool = True, use_ordered: bool = False):
+        """Generic streaming fit function for both index and timestamp keys.
+
+        This is a helper method that implements the common logic for streaming fits,
+        whether using batch indices or timestamps as keys.
+
+        :param batch_iter: Iterator over batches of problems.
+        :param key_fn: Function to extract key and batch from iterator items.
+        :param include_scores: Whether to include scores in the results (default: True).
+        :param use_ordered: Whether to use OrderedDict for results (default: False).
+
+        :returns: Mapping of keys to ranking results.
+        """
+        forecaster_data = {}
+        batch_results = OrderedDict() if use_ordered else {}
+
+        for i, item in enumerate(batch_iter):
+            key, batch = key_fn(i, item)
+            if self.verbose:
+                msg = f"Processing batch {key}" if not use_ordered else f"Processing batch {i} at {key}"
+                self.logger.debug(msg)
+
+            # Process each problem in the batch
+            for problem in batch:
+                all_probs, usernames = [], []
+                correct_option_idx = np.array(problem.correct_option_idx)
+                for forecast in problem.forecasts:
+                    username = forecast.username
+                    if username not in forecaster_data:
+                        forecaster_data[username] = []
+                    usernames.append(username)
+                    all_probs.append(forecast.unnormalized_probs)
+
+                # batch process the scores
+                all_probs = np.array(all_probs)
+                scores = self._score_fn(correct_option_idx, all_probs)
+
+                for username, score in zip(usernames, scores):
+                    forecaster_data[username].append(score)
+
+            # Generate rankings for this batch
+            batch_results[key] = forecaster_data_to_rankings(
+                forecaster_data, include_scores=include_scores, ascending=False, aggregate="mean"
+            )
+            if self.verbose:
+                log_ranking_table(self.logger, batch_results[key])
+
+        return batch_results
 
     def fit_stream(self, problem_iter: Iterator[List[ForecastProblem]], include_scores: bool = True) -> Dict[int, Tuple[Dict[str, Any], Dict[str, int]]]:
         """Fit the scoring rule to streaming problems and return incremental results.
@@ -158,40 +223,90 @@ class ScoringRule(ABC):
 
         :returns: Mapping of batch indices to ranking results.
         """
-        forecaster_data = {}
-        batch_results = {}
-        batch_id = 0
+        return self._fit_stream_generic(
+            problem_iter,
+            key_fn=lambda i, batch: (i, batch),
+            include_scores=include_scores,
+            use_ordered=False
+        )
 
-        for batch in problem_iter:
-            if self.verbose:
-                self.logger.debug(f"Processing batch {batch_id}")
-            for problem in batch:
-                correct_probs, all_probs, usernames = [], [], []
-                for forecast in problem.forecasts:
-                    username = forecast.username
-                    if username not in forecaster_data:
-                        forecaster_data[username] = []
-                    usernames.append(username)
-                    correct_probs.append(forecast.correct_prob)
-                    all_probs.append(forecast.probs)
+    def fit_stream_with_timestamp(self, problem_time_iter: Iterator[Tuple[str, List[ForecastProblem]]], include_scores: bool = True) -> OrderedDict:
+        """Fit the scoring rule to streaming problems with timestamps and return incremental results.
 
-                # batch process the scores
-                correct_probs = np.array(correct_probs)
-                all_probs = np.array(all_probs)
-                scores = self._score_fn(correct_probs, all_probs)
+        This method processes problems with associated timestamps and returns rankings
+        after each batch, maintaining chronological order.
 
-                for username, score in zip(usernames, scores):
-                    forecaster_data[username].append(score)
+        :param problem_time_iter: Iterator over (timestamp, problems) tuples.
+        :param include_scores: Whether to include scores in the results (default: True).
 
-            batch_results[batch_id] = forecaster_data_to_rankings(
-                forecaster_data, include_scores=include_scores, ascending=False, aggregate="mean")
+        :returns: Chronologically ordered mapping of timestamps to ranking results.
+        """
+        return self._fit_stream_generic(
+            problem_time_iter,
+            key_fn=lambda i, item: (item[0], item[1]),
+            include_scores=include_scores,
+            use_ordered=True
+        )
 
-            if self.verbose:
-                log_ranking_table(self.logger, batch_results[batch_id])
+    def fit_by_category(self, problems: List[ForecastProblem], include_scores: bool = True, stream_with_timestamp: bool = False,
+                        stream_increment_by: Literal["day", "week", "month"] = "day", min_bucket_size: int = 1) -> \
+            Tuple[Dict[str, Any], Dict[str, int]] | Dict[str, int]:
+        """Fit the scoring rule to the given problems by category.
 
-            batch_id += 1
+        This method processes problems grouped by category and returns rankings for each category.
+        Optionally, it can stream problems within each category over time.
 
-        return batch_results
+        :param problems: List of ForecastProblem instances to process.
+        :param include_scores: Whether to include scores in the results (default: True).
+        :param stream_with_timestamp: Whether to stream problems with timestamps (default: False).
+        :param stream_increment_by: The increment by which to stream problems (default: "day").
+        :param min_bucket_size: The minimum number of problems to include in a bucket (default: 1).
+
+        :returns: Mapping of categories to ranking results.
+        """
+        category_to_problems = dict()
+        for problem in problems:
+            if problem.category not in category_to_problems:
+                category_to_problems[problem.category] = []
+            category_to_problems[problem.category].append(problem)
+
+        if not stream_with_timestamp:
+            # simply fit the model to each category
+            results_dict = dict()
+            for category, category_problems in category_to_problems.items():
+                results_dict[category] = self.fit(category_problems, include_scores=include_scores)
+            results_dict["overall"] = self.fit(problems, include_scores=include_scores)
+            return results_dict
+        else:
+            # create a separate iterator for overall problems
+            overall_iterator = ForecastChallenge._stream_problems_over_time(
+                problems=problems,
+                increment_by=stream_increment_by,
+                min_bucket_size=min_bucket_size
+            )
+            # create a separate iterator for each category
+            results_dict = dict()
+            for category, category_problems in category_to_problems.items():
+                category_iterator = ForecastChallenge._stream_problems_over_time(
+                    problems=category_problems,
+                    increment_by=stream_increment_by,
+                    min_bucket_size=min_bucket_size
+                )
+
+                results_dict[category] = self._fit_stream_generic(
+                    category_iterator,
+                    key_fn=lambda i, item: (item[0], item[1]),
+                    include_scores=include_scores,
+                    use_ordered=True
+                )
+
+            results_dict["overall"] = self._fit_stream_generic(
+                overall_iterator,
+                key_fn=lambda i, item: (item[0], item[1]),
+                include_scores=include_scores,
+                use_ordered=True
+            )
+            return results_dict
 
 
 class LogScoringRule(ScoringRule):
@@ -217,21 +332,21 @@ class LogScoringRule(ScoringRule):
         self.logger.info(
             f"Initialized {self.__class__.__name__} with hyperparam: clip_prob={clip_prob}")
 
-    def _score_fn(self, correct_probs: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
+    def _score_fn(self, correct_option_idx: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
         """Calculate logarithmic scores for the forecasts.
 
         The logarithmic score is computed as log(p_correct), where p_correct is the
         predicted probability of the actual outcome. To prevent numerical issues,
         probabilities are clipped to a minimum value.
 
-        :param correct_probs: Array of predicted probabilities for the actual outcomes.
-                             Shape (n,) where n is the number of forecasts.
+        :param correct_option_idx: Array of indices of the correct options.
+                             Shape (m,) where m is the number of correct options.
         :param all_probs: Array of all predicted probability distributions.
                          Shape (n, k) where n is number of forecasts, k is number of options.
 
         :returns: Array of logarithmic scores. Shape (n,).
         """
-        return np.log(np.maximum(correct_probs, self.clip_prob))
+        return np.log(np.maximum(all_probs[:, correct_option_idx], self.clip_prob))
 
 
 class BrierScoringRule(ScoringRule):
@@ -259,19 +374,18 @@ class BrierScoringRule(ScoringRule):
         self.logger.info(
             f"Initialized {self.__class__.__name__} with hyperparam: negate={negate}")
 
-    def _score_fn(self, correct_probs: np.ndarray, all_probs: np.ndarray, negate: bool = True) -> np.ndarray:
+    def _score_fn(self, correct_option_idx: np.ndarray, all_probs: np.ndarray, negate: bool = True) -> np.ndarray:
         """Calculate Brier scores for the forecasts.
 
         The Brier score is computed as the average squared difference between
         predicted probabilities and actual outcomes. The formula is:
 
-        Brier Score = (1 - p_correct)² + Σ(p_incorrect)²
+        Brier Score = Σ(1 - p_correct)² for all correct options
 
-        where p_correct is the predicted probability of the actual outcome and
-        p_incorrect are the predicted probabilities of incorrect outcomes.
+        where p_correct is the predicted probability of the actual outcome.
 
-        :param correct_probs: Array of predicted probabilities for the actual outcomes.
-                             Shape (n,) where n is the number of forecasts.
+        :param correct_option_idx: Array of indices of the correct options.
+                             Shape (m,) where m is the number of correct options.
         :param all_probs: Array of all predicted probability distributions.
                          Shape (n, k) where n is number of forecasts, k is number of options.
         :param negate: Whether to negate the scores so that higher values are better
@@ -279,15 +393,16 @@ class BrierScoringRule(ScoringRule):
 
         :returns: Array of Brier scores. Shape (n,).
         """
-        # correct_probs is 1D with shape (n,), all_probs is 2D with shape (n, k)
-        # (1) we obtain (n,) correct_scores
-        correct_scores = (1 - correct_probs) ** 2 - correct_probs ** 2
-        # (2) we obtain (n,) incorrect scores
-        incorrect_scores = np.sum(all_probs ** 2, axis=1)
+        one_hot = np.zeros(all_probs.shape[1])
+        # correct_option_idx might be an empty array
+        if len(correct_option_idx) > 0:
+            one_hot[correct_option_idx] = 1
+        # ignore above
+        brier_scores = np.sum((all_probs - one_hot) ** 2, axis=1)
         # (3) we obtain (n,) scores, rescaled so that it lies in [0, 1]
-        scores = (correct_scores + incorrect_scores) / 2
+        scores = brier_scores / all_probs.shape[1]
         # (4) negate the result since higher scores are better
-        return -scores if negate else scores
+        return 1 - scores if negate else scores
 
 
 class SphericalScoringRule(ScoringRule):
@@ -308,7 +423,7 @@ class SphericalScoringRule(ScoringRule):
         super().__init__(verbose=verbose)
         self.logger.info(f"Initialized {self.__class__.__name__}")
 
-    def _score_fn(self, correct_probs: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
+    def _score_fn(self, correct_option_idx: np.ndarray, all_probs: np.ndarray) -> np.ndarray:
         """Calculate spherical scores for the forecasts.
 
         The spherical score is computed as the cosine similarity between the
@@ -319,22 +434,12 @@ class SphericalScoringRule(ScoringRule):
         where p_correct is the predicted probability of the actual outcome and
         $\\lVert p \\rVert$ is the L2 norm of the entire probability vector.
 
-        :param correct_probs: Array of predicted probabilities for the actual outcomes.
-                             Shape (n,) where n is the number of forecasts.
+        :param correct_option_idx: Array of indices of the correct options.
+                             Shape (m,) where m is the number of correct options.
         :param all_probs: Array of all predicted probability distributions.
                          Shape (n, k) where n is number of forecasts, k is number of options.
 
         :returns: Array of spherical scores. Shape (n,).
         """
         # formula: r_j / sum_i r_i where r_j is the correct probability of the j-th option
-        correct_scores = correct_probs / np.linalg.norm(all_probs, axis=1)
-        return correct_scores
-
-
-if __name__ == "__main__":
-    # check the implementation of the scoring rules
-    correct_probs = np.array([0, 0.5, 1])
-    all_probs = np.array([[0, 0.4, 0.6], [0.2, 0.5, 0.3], [0, 0, 1]])
-    print(LogScoringRule()._score_fn(correct_probs, all_probs))
-    print(BrierScoringRule()._score_fn(correct_probs, all_probs))
-    print(SphericalScoringRule()._score_fn(correct_probs, all_probs))
+        return np.sum(all_probs[:, correct_option_idx], axis=1) / np.linalg.norm(all_probs, axis=1)
